import os
import sys
import logging
from distlib.database import DistributionPath

TASKS_DIRECTORY = "firex_tasks_directory"


logger = logging.getLogger(__name__)


def _get_paths_without_cwd():
    # This is needed because Celery temporarily adds the cwd into the sys.path via a context switcher,
    # and our discovery takes place inside that context.
    # Having cwd in the sys.path can slow down the discovery significantly without any benefit.
    paths = list(sys.path)
    try:
        paths.remove(os.getcwd())
    except ValueError:  # pragma: no cover
        pass
    return paths


def _get_firex_dependant_package_locations()-> []:
    distributions = DistributionPath(path=_get_paths_without_cwd(), include_egg=True).get_distributions()

    # some packages (such as any tree) might cause exceptions in logging
    old_raise = logging.raiseExceptions
    try:
        logging.raiseExceptions = False
        firex_app_name = __name__.split(".")[0]
        logging.getLogger('distlib.metadata').setLevel(logging.WARNING)
        logging.getLogger('distlib.database').setLevel(logging.WARNING)
        dependants = []
        while True:
            try:
                d = next(distributions)
            except StopIteration:
                break
            except Exception as e:
                logger.debug(f"Failed to get a distribution due to {e}")
            else:
                if firex_app_name in d.run_requires:
                    dependants.append(d)
    finally:
        logging.raiseExceptions = old_raise

    locations = []
    for d in dependants:
        top = os.path.join(d.path, "top_level.txt")
        with open(top) as t:
            top_package = t.read().strip()
            package_location = os.path.join(os.path.dirname(d.path), top_package)
            locations.append(package_location)
    return locations


def discover_package_modules(current_path, root_path=None) -> []:
    if root_path is None:
        root_path = os.path.dirname(current_path)

    services = []
    if os.path.isfile(current_path):
        basename, ext = os.path.splitext(current_path)
        if ext.lower() == ".py" and not os.path.basename(current_path).startswith('_'):
            basename = basename.replace(root_path, "")
            return [basename.replace(os.path.sep, ".").strip(".")]
        else:
            return []
    elif os.path.isdir(current_path):
        base = os.path.basename(current_path)
        if "__pycache__" in base or base.startswith("."):
            return []
        for child_name in os.listdir(current_path):
            full_child = os.path.join(current_path, child_name)
            services += discover_package_modules(full_child, root_path)
        return services
    else:
        # either a symlink or a path that doesn't exist
        return []


def find_firex_task_bundles()->[]:
    # look for task modules in dependant packages
    bundles = []
    for location in _get_firex_dependant_package_locations():
        bundles += discover_package_modules(location)

    # look for task modules in env defined location
    if TASKS_DIRECTORY in os.environ:
        include_location = os.environ[TASKS_DIRECTORY]
        if os.path.isdir(include_location):
            if include_location not in sys.path:
                sys.path.append(include_location)
            include_tasks = discover_package_modules(include_location, root_path=include_location)
            bundles += include_tasks

    return bundles



from getpass import getuser
import time

from firexkit.argument_conversion import SingleArgDecorator
from firexkit.task import FireXTask

from firexapp.engine.celery import app
from firexapp.submit.arguments import InputConverter


@app.task
def nop():
    return


@app.task
def sleep(sleep=None):
    if sleep:
        time.sleep(int(sleep))
    return


@app.task(returns='username')
def getusername():
    return getuser()


# The @app.task() makes this normal python function a FireX Service.
@app.task(returns=['greeting'])
def greet(name=getuser()):
    assert len(name) > 1, "Cannot greet a name with 1 or fewer characters."
    return 'Hello %s!' % name


# Setting bind=True makes the first argument received by the service 'self'. It's most commonly used to invoke
# (enqueue) other services, but provides much more functionality as outlined here:
@app.task(bind=True, returns=['guests_greeting'])
def greet_guests(self: FireXTask, guests):
    child_promises = []
    for guest in guests:
        # Create a Celery Signature, see: https://docs.celeryproject.org/en/latest/userguide/canvas.html#signatures
        greet_signature = greet.s(name=guest)
        # enqueue_child by default schedules the supplied signature for execution asynchronously and immediately
        # returns the newly created child result promise.
        child_promise = self.enqueue_child(greet_signature)
        child_promises.append(child_promise)

    # We want to get all the return values (greetings) from child tasks, but we must wait first to make sure they're all
    # available before inspecting them with child_promise.result[<returns_key>]
    self.wait_for_children(raise_exception_on_failure=False)
    # Since wait_for_children has completed, we know it's safe to inspect the results of all child task promises,
    # after verifying the task was a success.
    greetings = [promise.result['greeting'] for promise in child_promises if promise.successful()]

    if any(promise.failed() for promise in child_promises):
        greetings.append("And apologies to those not mentioned.")

    return ' '.join(greetings)


@InputConverter.register
@SingleArgDecorator('guests')
def to_list(guests):
    return guests.split(',')


@app.task(returns=['amplified_message'])
def amplify(to_amplify):
    return to_amplify.upper()


@app.task(bind=True, returns=['amplified_greeting'])
def amplified_greet_guests(self: FireXTask, guests):

    # Create a chain that can be enqueued. The greet_guests service will produce a guests_greeting,
    # which will then be delivered to amplify as its to_amplify argument.
    amplified_greet_guests_chain = greet_guests.s(guests=guests) | amplify.s(to_amplify='@guests_greeting')

    # Chains can be enqueued just like signatures. You can consider a signature a chain with only one service.
    chain_results = self.enqueue_child_and_get_results(amplified_greet_guests_chain)
    return chain_results['amplified_message']



from abc import ABC, abstractmethod
from collections import OrderedDict
from firexapp.submit.install_configs import FireXInstallConfigs

TRACKING_SERVICE_ENTRY_POINT = 'firex_tracking_service'

_services = None


#
# In case there are duplicate modules found, only keep one for each
#   (name, module_name, object_name) tuple. This prevents duplicate
#   arg registration failures when the sys.path causes the same service
#   to be found twice.
#
def prune_duplicate_module_entry_points(entry_points):
    id_to_entry_points = OrderedDict()

    for e in entry_points:
        key = (e.name, e.module_name, e.object_name)
        if key not in id_to_entry_points:
            id_to_entry_points[key] = e
        # Replace the currently stored entry point for this key if the distro is None.
        elif id_to_entry_points[key].distro is None and e.distro is not None:
            id_to_entry_points[key] = e

    return list(id_to_entry_points.values())


def get_tracking_services() -> ():
    global _services
    if _services is None:
        import entrypoints
        entry_pts = [entry_point for entry_point in entrypoints.get_group_all(TRACKING_SERVICE_ENTRY_POINT)]
        entry_pts = prune_duplicate_module_entry_points(entry_pts)
        entry_objects = [e.load() for e in entry_pts]
        _services = tuple([point() for point in entry_objects])
    return _services


def has_flame():
    # Unfortunate coupling, but just too many things vary depending on presence of flame. Will eventually bring
    # flame in to firexapp.
    return 'FlameLauncher' in get_tracking_services()


class TrackingService(ABC):

    install_configs: FireXInstallConfigs

    def extra_cli_arguments(self, arg_parser):
        pass

    @abstractmethod
    def start(self, args, install_configs: FireXInstallConfigs, **kwargs) -> {}:
        self.install_configs = install_configs

    def ready_for_tasks(self, **kwargs) -> bool:
        return True

    def ready_release_console(self, **kwargs) -> bool:
        return True


def get_service_name(service: TrackingService) -> str:
    return service.__class__.__name__
